YOUSPEED = 7
SLIMESPEEDMIN = 2
SLIMESPEEDMAX = 6
LEADDISTANCE = 1
LEADATTACHDISTANCE = 3
HEALTHFACTOR = 0.03
WATERCONSUMPTIONAMOUNT = 0.1
WATERDECREASE = 0.02
WATERINCREASEFACTOR = 0.2
FOODCONSUMPTIONAMOUNT = 0.1
FOODDECREASE = 0.02
FOODINCREASEFACTOR = 0.2
WATERANIMATIONLENGTH = 15
SLIMEWABBLESPEED = 0.5

mod = math.fmod
if mod == nil then
  mod = math.mod
end

function dump(o)
  return dump_impl(o, 0)
end
function dump_impl(o, i)
   if type(o) == 'table' then
      local s = string.rep('    ', i)..'{\n'
      for k,v in pairs(o) do
         if type(k) ~= 'number' then k = '"'..k..'"' end
         s = s .. string.rep('    ', i)..'  ['..k..'] = ' .. dump_impl(v, i + 1) .. ',\n'
      end
      return s .. string.rep("    ", i)..'}\n'
   else
      return tostring(o)
   end
end

function reset()
  died = false
  youx = -1
  youy = 0
  slimes = {{["x"] = 1, ["y"] = 0, ["size"] = 0.5, ["water"] = 1, ["food"] = 1, ["phase5"] = math.random(2*math.pi), ["phase7"] = math.random(2*math.pi)}}
  waters = {}
  foods = {}
  tfood = math.random(0, 2) + 2
  twater = math.random(0, 2) + 2
  slimeonlead = 1
  leadlocked = true
  totalhealth = 0.5
  maxtotalhealth = 0.5
  start = love.timer.getTime()
end

function watertexturemap(x, y, r, g, b, a)
  sy = (1+math.sin(2*math.pi*y/WATERANIMATIONLENGTH))/4
  return sy/2, sy/2, 1
end

function calculatecoordinates()
  text = love.graphics.newText(font, string.format("Time:\n%.0f\n\nTotal Health:\n%.2f\n\nMax Total Health:\n%.2f\n\nCurrent Slime:\n%.1f%% Health\n%.1f%% Water\n%.1f%% Food",9999, 99.99,99.99, 99.9, 99.9, 199.9, 99.9))
  textwidth = text:getWidth() + 20
  textheight = text:getHeight()
  scale = math.max(0, math.min((screenheight - 20)/(2*halfheight + 2), (screenwidth - 30 - textwidth)/(2*halfwidth + 2)))
  cx=(screenwidth + (halfwidth + 1)*scale - textwidth)/3
  cy=screenheight/2
  ix = 2 * cx
  slimevertexcount = 30
  slimemesh = love.graphics.newMesh(slimevertexcount + 2, "fan", "stream")
  slimemesh:setVertex(1, 0, 0, 0, 0, 0.5, 0.5, 0.5, 1)
  watertexturedata = love.image.newImageData(2*scale + 1, 2*scale + WATERANIMATIONLENGTH + 1)
  watertexturedata:mapPixel(watertexturemap)
  watertexture = love.graphics.newImage(watertexturedata)
end

function love.load()
  font=love.graphics.setNewFont(24)
  fontheight=font:getHeight()
  screenwidth=love.graphics.getWidth()
  screenheight=love.graphics.getHeight()
  halfwidth = 10
  halfheight = 10
  calculatecoordinates()
  math.randomseed(os.time())
  reset()
  help = true
  start = 0
  highscoretime = 0
  highscoretotalhealth = 0
  scores = love.filesystem.read("SlimePetsHighScores.txt")
  if scores ~= nil then
    highscoretime = tonumber(string.sub(scores, 1, 10))
    highscoretotalhealth = tonumber(string.sub(scores, 11, 20))
    print("'"..scores.."'"..string.sub(scores, 1, 10).."'"..string.sub(scores, 11, 20).."'"..highscoretime.."'"..highscoretotalhealth.."'")
    if highscoretime == nil then
      highscoretime = 0
    end
    if highscoretotalhealth == nil then
      highscoretotalhealth = 0
    end
  end
end

function love.resize()
  screenwidth=love.graphics.getWidth()
  screenheight=love.graphics.getHeight()
  calculatecoordinates()
end

function love.update(dt)
  if died or help then
    return
  end
  for i, water in ipairs(waters) do
    water["phase"] = mod(water["phase"] + dt*5, WATERANIMATIONLENGTH)
    water["angle"] = mod(water["angle"] + dt/50, 2*math.pi)
  end
  twater = twater - dt
  tfood = tfood - dt
  if twater < 0 then
    waters[#waters + 1] = {["x"] = math.random(-halfwidth, halfwidth), ["y"] = math.random(-halfheight, halfheight), ["size"] = 1, ["angle"] = math.random(2*math.pi), ["phase"] = 0}
    twater = twater + 10 +  math.random(0, 20)
  end
  if tfood < 0 then
    foods[#foods + 1] = {["x"] = math.random(-halfwidth, halfwidth), ["y"] = math.random(-halfheight, halfheight), ["size"] = 1}
    tfood = tfood + 10 + math.random(0, 20)
  end
  newslimes = {}
  totalhealth = 0
  for i, slime in ipairs(slimes) do
    watercount = 0
    foodcount = 0
    todelete = {}
    for j, water in ipairs(waters) do
      deltax = slime["x"] - water["x"]
      deltay = slime["y"] - water["y"]
      distance = math.sqrt(deltax*deltax + deltay*deltay)
      if distance < slime["size"] + water["size"] then
        watercount = watercount + 1
        water["size"] = water["size"] - WATERCONSUMPTIONAMOUNT*dt
        if water["size"] < 0 then
          todelete[#todelete + 1] = j
        end
      end
    end
    deleted = 0
    for _, j in ipairs(todelete) do
      table.remove(waters, j - deleted)
      deleted = deleted + 1
    end
    todelete = {}
    for j, food in ipairs(foods) do
      deltax = slime["x"] - food["x"]
      deltay = slime["y"] - food["y"]
      distance = math.sqrt(deltax*deltax + deltay*deltay)
      if distance < slime["size"] + food["size"] then
        foodcount = foodcount + 1
        food["size"] = food["size"] - FOODCONSUMPTIONAMOUNT*dt
        if food["size"] < 0 then
          todelete[#todelete + 1] = j
        end
      end
    end
    deleted = 0
    for _, j in ipairs(todelete) do
      table.remove(foods, j - deleted)
      deleted = deleted + 1
    end
    slime["water"] = math.max(0, math.min(2, slime["water"] + (WATERINCREASEFACTOR*watercount - WATERDECREASE)*dt))
    slime["food"] = math.max(0, math.min(1, slime["food"] + (FOODINCREASEFACTOR*foodcount - FOODDECREASE)*dt))
    slime["size"] = math.max(0, math.min(1, slime["size"] + dt*HEALTHFACTOR*math.min(2 * slime["food"] - 1, 1 - 2 * math.abs(slime["water"] - 1))))
    slime["phase5"] = mod(slime["phase5"] + 3*SLIMEWABBLESPEED*dt, 2*math.pi)
    slime["phase7"] = mod(slime["phase7"] + 2*SLIMEWABBLESPEED*dt, 2*math.pi)
    totalhealth = totalhealth + slime["size"]
    if slime["size"] == 0 then
      if not died then
        died = true
        start = love.timer.getTime() - start
        savescores = false
        if start > highscoretime then
          highscoretime = start
          savescores = true
        end
        if maxtotalhealth > highscoretotalhealth then
          highscoretotalhealth = maxtotalhealth
          savescores = true
        end
        if savescores then
          love.filesystem.write("SlimePetsHighScores.txt", string.format("%10.2f%10.2f", highscoretime, highscoretotalhealth))
        end
      end
    elseif slime["size"] == 1 then
      angle = math.random(2*math.pi)
      deltax = math.sin(angle)/2
      deltay = math.cos(angle)/2
      newslimes[#newslimes + 1] = {["x"] = slime["x"] + deltax, ["y"] = slime["y"]+deltay, ["size"] = 0.5, ["water"] = slime["water"], ["food"] = slime["food"], ["phase5"] = math.random(2*math.pi), ["phase7"] = math.random(2*math.pi)}
      slime["size"] = 0.5
      slime["x"] = slime["x"] - deltax
      slime["y"] = slime["y"] - deltay
    end
  end
  for i, slime in ipairs(newslimes) do
    slimes[#slimes + 1]  = slime
  end
  if totalhealth > maxtotalhealth then
    maxtotalhealth = totalhealth
  end
  deltax = (love.mouse.getX() - cx)/scale - youx
  deltay = (love.mouse.getY() - cy)/scale - youy
  speed = YOUSPEED*dt
  distance = math.sqrt(deltax*deltax + deltay*deltay)
  if distance > speed then
    youx = math.max(-halfheight, math.min(halfwidth, youx + deltax*speed/distance))
    youy = math.max(-halfheight, math.min(halfheight, youy + deltay*speed/distance))
  else
    youx = youx + deltax
    youy = youy + deltay
  end
  if leadlocked then
    slime = slimes[slimeonlead]
    deltax = youx - slime["x"]
    deltay = youy - slime["y"]
    speed = (SLIMESPEEDMIN + (SLIMESPEEDMAX - SLIMESPEEDMIN) * slime["size"])*dt
    distance = math.sqrt(deltax*deltax + deltay*deltay)
    if distance > LEADDISTANCE + speed then
      slime["x"] = slime["x"] + deltax*speed/distance
      slime["y"] = slime["y"] + deltay*speed/distance
    elseif distance > LEADDISTANCE then
      slime["x"] = youx - LEADDISTANCE*deltax/distance
      slime["y"] = youy - LEADDISTANCE*deltay/distance
    end
  else
    slimeonlead = nil
    closestDistance = nil
    for i, slime in ipairs(slimes) do
      deltax = youx - slime["x"]
      deltay = youy - slime["y"]
      distance = math.sqrt(deltax*deltax + deltay*deltay)
      if distance < LEADATTACHDISTANCE and (closestDistance == nil or closestDistance > distance) then
        closestDistance = distance
        slimeonlead = i
      end
    end
  end
end

function love.keypressed(key)
  if key == "f1" or key == "p"  or key == "h" then
    if help then 
      help = false
    else
      help = true
    end
    if not died then
      start = love.timer.getTime() - start
    end
  elseif key == "space" then
    if died or help then
      return
    end
    if leadlocked then
      leadlocked = false
    elseif slimeonlead ~= nil then
      leadlocked = true
    end
  elseif key == "r" and died and not help then
    reset()
  end
end

function love.draw()
  if help then
    love.graphics.setColor(1, 1, 1)
    love.graphics.printf("Pet Slime! The latest craze sweeping the world!\n\nYou have just been given a new pet slime, try and keep it (them) alive. \n\nYou can use your lead to move the slime around to the water and food they need. If they have too much or too little water (blue) or not enough food (grey) they will shrivel (while purple) and eventually die. However if you can keep them happy they will grow and grow (while yellow), oh, and then split into two slimes that you now need to care for.\n\nTry and keep your slimes alive for as long as possible or try and get the best set of slimes (most total health) but if any of them die it's game over.\n\nGood luck and have fun!\n\nUse the mouse to move, press 'space' to attach or remove the lead and 'p' or 'h' to show or hide this help message.",10,10,screenwidth-20,"center")
  else
    love.graphics.setStencilTest("greater", 0)
    for i, water in ipairs(waters) do
      x = cx + scale*water["x"]
      y = cy + scale*water["y"]
      r = scale*water["size"]
      love.graphics.setColor(1, 1, 1)
      love.graphics.stencil(function() love.graphics.circle("fill", x, y, r) end)
      love.graphics.rectangle("fill", 0, 0, screenwidth, screenheight)
      love.graphics.draw(watertexture, x, y, water["angle"], 1, 1, r, r + water["phase"])
    end
    love.graphics.setStencilTest()
    for i, food in ipairs(foods) do
      love.graphics.setColor(0.5, 0.5, 0.5)
      love.graphics.circle("fill", cx + scale*food["x"], cy + scale*food["y"], scale*food["size"])
    end
    if slimeonlead ~= nil then
      slime = slimes[slimeonlead]
      if leadlocked then
        love.graphics.setColor(1, 1, 1)
      else
        love.graphics.setColor(0.5, 0.5, 0.5)
      end
      love.graphics.line(cx + scale*youx, cy + scale*youy, cx + scale*slime["x"], cy + scale*slime["y"])
    end
    for i, slime in ipairs(slimes) do
      health = math.min(2 * slime["food"] - 1, 1 - 2 * math.abs(slime["water"] - 1))
      meshdata = {}
      for i=1,slimevertexcount do
        theta = 2*math.pi*i/slimevertexcount
        r = scale*(slime["size"] + (math.sin(5*theta + slime["phase5"]) + math.sin(7*theta + slime["phase7"]))/20)
        meshdata[i] = {r*math.cos(theta), r*math.sin(theta), 0, 0, 1, 1, 1, 1}
      end
      meshdata[slimevertexcount + 1] = meshdata[1]
      slimemesh:setVertices(meshdata, 2)
      love.graphics.setColor(0.3 + health*0.7,  health, 0.3 - health*0.3)
      love.graphics.draw(slimemesh, cx + scale*slime["x"], cy + scale*slime["y"])
    end
    love.graphics.setColor(1, 1, 1)
    love.graphics.circle("fill", cx + scale*youx, cy + scale*youy, scale*0.2)
    love.graphics.setColor(0.1,0.1,0.1)
    love.graphics.rectangle("fill", ix, 10, textwidth, screenheight - 20)
    love.graphics.setColor(1, 1, 1)
    water = nil
    food = nil
    now = love.timer.getTime() - start
    if died then
      text = love.graphics.newText(font, string.format("Time:\n%.0f\n\nTotal Health:\n%.2f\n\nMax Total Health:\n%.2f\n\nHigh Scores:\nTime:\n%.0f\n\nMax Total Health:\n%.2f", start, totalhealth, maxtotalhealth, highscoretime, highscoretotalhealth))
    elseif slimeonlead ~= nil then
      text = love.graphics.newText(font, string.format("Time:\n%.0f\n\nTotal Health:\n%.2f\n\nMax Total Health:\n%.2f\n\nCurrent Slime:\n%.1f%% Health\n%.1f%% Water\n%.1f%% Food", now, totalhealth, maxtotalhealth, slimes[slimeonlead]["size"]*100, slimes[slimeonlead]["water"]*100, slimes[slimeonlead]["food"]*100))
    else
      text = love.graphics.newText(font, string.format("Time:\n%.0f\n\nTotal Health:\n%.2f\n\nMax Total Health:\n%.2f\n\nCurrent Slime:\nNone\nNone\nNone", now, totalhealth,maxtotalhealth))
    end
    love.graphics.draw(text, ix + 10, cy - textheight/2)
    if died then
      love.graphics.printf("Oh dear a slime died!\n\nPress 'r' to try again.",5,cy,screenwidth-10,"center")
    end
  end
end